@@ -1019,8 +1019,20 @@ def InverseDynamics(thetalist, dthetalist, ddthetalist, g, Ftip, Mlist, \
     return taulist
 '''
 def InverseDynamics(thetalist, dthetalist, ddthetalist, g, Ftip, Mlist, \
-                    Glist, Slist,eef_mass = 0 ):
-    n = len(thetalist)
+                    Glist, Slist,eef_mass = 0 ,eef_inertia= np.zeros((3,3))):
+    n = len(thetalist)    
+    thetalist=np.append(thetalist,0)
+    dthetalist=np.append(dthetalist,0)
+    ddthetalist=np.append(ddthetalist,0)    
+    Meef = np.eye(4);
+    Mlist=np.concatenate((Mlist,[Meef]),axis=0)
+    Geef=np.eye(6)*eef_mass;
+    Geef[0:3,0:3] = eef_inertia;
+    Glist=np.concatenate((Glist,[Geef]),axis=0)
+    Slist=np.concatenate((Slist,np.zeros((6, 1))),axis=1)
+
+
+    n = n+1
     Mi = np.eye(4)
     Ai = np.zeros((6, n))
     AdTi = [[None]] * (n + 1)
@@ -1030,11 +1042,6 @@ def InverseDynamics(thetalist, dthetalist, ddthetalist, g, Ftip, Mlist, \
     AdTi[n] = Adjoint(TransInv(Mlist[n]))
     Fi = np.array(Ftip).copy()
     taulist = np.zeros(n)
-    #Glist[-1] = Glist[-1]+  
-    Geef=np.eye(6)*eef_mass;
-    Geef[0:3,0:3] = 0;
-    Meef = np.eye(4);
-
     for i in range(n):
         Mi = np.dot(Mi,Mlist[i])
         Ai[:, i] = np.dot(Adjoint(TransInv(Mi)), np.array(Slist)[:, i])
@@ -1045,32 +1052,18 @@ def InverseDynamics(thetalist, dthetalist, ddthetalist, g, Ftip, Mlist, \
         Vdi[:, i + 1] = np.dot(AdTi[i], Vdi[:, i]) \
                        + Ai[:, i] * ddthetalist[i] \
                        + np.dot(ad(Vi[:, i + 1]), Ai[:, i]) * dthetalist[i]
-    Aeef = np.dot(Adjoint(TransInv(Meef)), np.array(Slist)[:, -1])
-    AdTeef = Adjoint(np.dot(MatrixExp6(VecTose3(Aeef * \
-                                            -0)), \
-                                 TransInv(Meef)))
-    Veef= np.dot(AdTeef, Vi[:,-1]) + Ai[:, -1] * 0
-    Vdeef = np.dot(AdTeef, Vdi[:,-1]) \
-	               +  Ai[:, -1]* 0\
-	               + np.dot(ad(Veef), Ai[:, -1]) * 0                       
     for i in range (n - 1, -1, -1):
-	    if i==n-1:
-	    	Fi =np.dot(np.array(AdTeef).T, Fi) \
-		 + np.dot(np.array(Geef), Vdeef) \
-		 - np.dot(np.array(ad(Veef)).T, \
-		          np.dot(np.array(Geef), Veef))    
-	    Fi = np.dot(np.array(AdTi[i + 1]).T, Fi) \
-		 + np.dot(np.array(Glist[i]), Vdi[:, i + 1]) \
-		 - np.dot(np.array(ad(Vi[:, i + 1])).T, \
-		          np.dot(np.array(Glist[i]), Vi[:, i + 1]))
-
-	    taulist[i] = np.dot(np.array(Fi).T, Ai[:, i])
-	    
+        Fi = np.dot(np.array(AdTi[i + 1]).T, Fi) \
+             + np.dot(np.array(Glist[i]), Vdi[:, i + 1]) \
+             - np.dot(np.array(ad(Vi[:, i + 1])).T, \
+                      np.dot(np.array(Glist[i]), Vi[:, i + 1]))
+        taulist[i] = np.dot(np.array(Fi).T, Ai[:, i])
     return taulist
+
     
     
 
-def MassMatrix(thetalist, Mlist, Glist, Slist):
+def MassMatrix(thetalist, Mlist, Glist, Slist,eef_mass=0,eef_inertia= np.zeros((3,3))):
     """Computes the mass matrix of an open chain robot based on the given
     configuration
 
@@ -1123,12 +1116,13 @@ def MassMatrix(thetalist, Mlist, Glist, Slist):
     for i in range (n):
         ddthetalist = [0] * n
         ddthetalist[i] = 1
-        M[:, i] = InverseDynamics(thetalist, [0] * n, ddthetalist, \
+        val = InverseDynamics(thetalist, [0] * n, ddthetalist, \
                                   [0, 0, 0], [0, 0, 0, 0, 0, 0], Mlist, \
-                                  Glist, Slist)
+                                  Glist, Slist,eef_mass,eef_inertia)
+        M[:, i] = val[0:n]
     return M
 
-def VelQuadraticForces(thetalist, dthetalist, Mlist, Glist, Slist):
+def VelQuadraticForces(thetalist, dthetalist, Mlist, Glist, Slist,eef_mass=0,eef_inertia= np.zeros((3,3))):
     """Computes the Coriolis and centripetal terms in the inverse dynamics of
     an open chain robot
 
@@ -1173,11 +1167,13 @@ def VelQuadraticForces(thetalist, dthetalist, Mlist, Glist, Slist):
     Output:
         np.array([0.26453118, -0.05505157, -0.00689132])
     """
-    return InverseDynamics(thetalist, dthetalist, [0] * len(thetalist), \
+    val = InverseDynamics(thetalist, dthetalist, [0] * len(thetalist), \
                            [0, 0, 0], [0, 0, 0, 0, 0, 0], Mlist, Glist, \
-                           Slist)
+                           Slist,eef_mass,eef_inertia)
+    n = len(thetalist);
+    return val[0:n]
 
-def GravityForces(thetalist, g, Mlist, Glist, Slist):
+def GravityForces(thetalist, g, Mlist, Glist, Slist,eef_mass=0,eef_inertia= np.zeros((3,3))):
     """Computes the joint forces/torques an open chain robot requires to
     overcome gravity at its configuration
 
@@ -1223,10 +1219,11 @@ def GravityForces(thetalist, g, Mlist, Glist, Slist):
         np.array([28.40331262, -37.64094817, -5.4415892])
     """
     n = len(thetalist)
-    return InverseDynamics(thetalist, [0] * n, [0] * n, g, \
-                           [0, 0, 0, 0, 0, 0], Mlist, Glist, Slist)
+    val = InverseDynamics(thetalist, [0] * n, [0] * n, g, \
+                           [0, 0, 0, 0, 0, 0], Mlist, Glist, Slist,eef_mass,eef_inertia)
+    return val[0:n]
 
-def EndEffectorForces(thetalist, Ftip, Mlist, Glist, Slist):
+def EndEffectorForces(thetalist, Ftip, Mlist, Glist, Slist,eef_mass=0,eef_inertia= np.zeros((3,3))):
     """Computes the joint forces/torques an open chain robot requires only to
     create the end-effector force Ftip
 
@@ -1274,7 +1271,7 @@ def EndEffectorForces(thetalist, Ftip, Mlist, Glist, Slist):
     """
     n = len(thetalist)
     return InverseDynamics(thetalist, [0] * n, [0] * n, [0, 0, 0], Ftip, \
-                           Mlist, Glist, Slist)
+                           Mlist, Glist, Slist,eef_mass,eef_inertia)
 
 def ForwardDynamics(thetalist, dthetalist, taulist, g, Ftip, Mlist, \
                     Glist, Slist):
@@ -1365,7 +1362,7 @@ def EulerStep(thetalist, dthetalist, ddthetalist, dt):
            dthetalist + dt * np.array(ddthetalist)
 
 def InverseDynamicsTrajectory(thetamat, dthetamat, ddthetamat, g, \
-                              Ftipmat, Mlist, Glist, Slist):
+                              Ftipmat, Mlist, Glist, Slist,eef_mass=0,eef_inertia= np.zeros((3,3))):
     """Calculates the joint forces/torques required to move the serial chain
     along the given trajectory using inverse dynamics
 
@@ -1462,7 +1459,7 @@ def InverseDynamicsTrajectory(thetamat, dthetamat, ddthetamat, g, \
         taumat[:, i] \
         = InverseDynamics(thetamat[:, i], dthetamat[:, i], \
                           ddthetamat[:, i], g, Ftipmat[:, i], Mlist, \
-                          Glist, Slist)
+                          Glist, Slist,eef_mass,eef_inertia)
     taumat = np.array(taumat).T
     return taumat
 
@@ -1708,7 +1705,52 @@ def JointTrajectory(thetastart, thetaend, Tf, N, method):
         traj[:, i] = s * np.array(thetaend) + (1 - s) * np.array(thetastart)
     traj = np.array(traj).T
     return traj
-
+def TransToTwist(Xd_list,dt):
+    n = len(Xd_list)
+    prevXd =np.reshape(Xd_list[0,:,:],[4,4])
+    Vd_list = []
+    for i in range(0,n):
+        Xd = np.reshape(Xd_list[i,:,:],[4,4])
+        Vd = se3ToVec(MatrixLog6(TransInv(prevXd)@Xd))/dt;
+        prevXd = Xd;
+        Vd_list.append(Vd);
+    return np.array(Vd_list);
+def TransToVel(Xd_list,dt):
+    n = len(Xd_list)
+    prevXd =np.reshape(Xd_list[0,:,:],[4,4])
+    Vd_list = []
+    for i in range(0,n):
+        Xd = np.reshape(Xd_list[i,:,:],[4,4])
+        Rd,pd = TransToRp(Xd)
+        prevRd,prevpd = TransToRp(prevXd)
+        w = so3ToVec(MatrixLog3(prevRd.T @ Rd ))/dt
+        v = (pd - prevpd)/dt
+        Vd = np.r_[w,v];
+        prevXd = Xd;
+        Vd_list.append(Vd);
+    return np.array(Vd_list);
+def TransToAcc(Xd_list,dt):
+    n = len(Xd_list)
+    Vd_list = TransToVel(Xd_list,dt)
+    dVd_list  =[]
+    prevVd =Vd_list[0,:]
+    for i in range(0,n):
+        Vd = Vd_list[i,:];
+        dVd = (Vd-prevVd)/dt
+        prevVd = Vd;
+        dVd_list.append(dVd);
+    return np.array(dVd_list);
+def TransToTwistDot(Xd_list,dt):
+    n = len(Xd_list)
+    Vd_list = TransToTwist(Xd_list,dt)
+    prevVd = Vd_list[0,:];
+    dVd_list = []
+    for i in range(0,n):
+        Vd = Vd_list[i,:];
+        dVd = (Vd-prevVd)/dt
+        prevVd = Vd;
+        dVd_list.append(dVd);
+    return np.array(dVd_list);
 def ScrewTrajectory(Xstart, Xend, Tf, N, method,type=0):
     """Computes a trajectory as a list of N SE(3) matrices corresponding to
       the screw motion about a space screw axis
@@ -1870,7 +1912,7 @@ def CartesianTrajectory(Xstart, Xend, Tf, N, method,type=0):
 '''
 
 def ComputedTorque(thetalist, dthetalist, eint, g, Mlist, Glist, Slist, \
-                   thetalistd, dthetalistd, ddthetalistd, Kp, Ki, Kd):
+                   thetalistd, dthetalistd, ddthetalistd, Kp, Ki, Kd,eef_mass=0,eef_inertia=np.zeros((3,3))):
     """Computes the joint control torques at a particular time instant
 
     :param thetalist: n-vector of joint variables
@@ -1934,7 +1976,7 @@ def ComputedTorque(thetalist, dthetalist, eint, g, Mlist, Glist, Slist, \
                   Kp * e + Ki * (np.array(eint) + e) \
                   + Kd * np.subtract(dthetalistd, dthetalist)) \
            + InverseDynamics(thetalist, dthetalist, ddthetalistd, g, \
-                             [0, 0, 0, 0, 0, 0], Mlist, Glist, Slist)
+                             [0, 0, 0, 0, 0, 0], Mlist, Glist, Slist,eef_mass,eef_inertia)
 
 def SimulateControl(thetalist, dthetalist, g, Ftipmat, Mlist, Glist, \
                     Slist, thetamatd, dthetamatd, ddthetamatd, gtilde, \
